"""
Implement KeywordSearches.

This is a static library of generally-useful code for searching data based on
pre-defined keywords (in the programming language sense).

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
from typing import Any, Dict, Generator, List

from ruamel.yaml.comments import CommentedMap

from yamlpath.types import AncestryEntry, PathSegment
from yamlpath.enums import PathSearchKeywords, PathSearchMethods
from yamlpath.common import Anchors, Nodes, Searches
from yamlpath.path import SearchKeywordTerms
from yamlpath.exceptions import YAMLPathException
from yamlpath.wrappers import NodeCoords
from yamlpath import YAMLPath

class KeywordSearches:
    """Helper methods for common data searching operations."""

    @staticmethod
    def search_matches(
        terms: SearchKeywordTerms, haystack: Any, yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Perform a keyword search.

        Parameters:
        1. terms (SearchKeywordTerms) The search operation to perform
        2. haystack (Any) The data to evaluate
        3. yaml_path (YAMLPath) YAML Path containing this search keyword

        Keyword Arguments:  See each of the called KeywordSearches methods

        Returns:  (Generator[NodeCoords, None, None]) Matching data as it is
            generated
        """
        invert: bool = terms.inverted
        keyword: PathSearchKeywords = terms.keyword
        parameters: List[str] = terms.parameters
        nc_matches: Generator[NodeCoords, None, None]

        if keyword is PathSearchKeywords.HAS_CHILD:
            nc_matches = KeywordSearches.has_child(
                haystack, invert, parameters, yaml_path, **kwargs)
        elif keyword is PathSearchKeywords.NAME:
            nc_matches = KeywordSearches.name(
                invert, parameters, yaml_path, **kwargs)
        elif keyword is PathSearchKeywords.MAX:
            nc_matches = KeywordSearches.max(
                haystack, invert, parameters, yaml_path, **kwargs)
        elif keyword is PathSearchKeywords.MIN:
            nc_matches = KeywordSearches.min(
                haystack, invert, parameters, yaml_path, **kwargs)
        elif keyword is PathSearchKeywords.PARENT:
            nc_matches = KeywordSearches.parent(
                haystack, invert, parameters, yaml_path, **kwargs)
        else:
            raise YAMLPathException(
                "Unsupported search keyword {} in".format(keyword),
                str(yaml_path))

        for nc_match in nc_matches:
            yield nc_match

    @staticmethod
    # pylint: disable=locally-disabled,too-many-locals
    def has_child(
        data: Any, invert: bool, parameters: List[str], yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Indicate whether data has a named or anchored child.

        Parameters:
        1. data (Any) The data to evaluate
        2. invert (bool) Invert the evaluation
        3. parameters (List[str]) Parsed parameters
        4. yaml_path (YAMLPath) YAML Path begetting this operation

        Keyword Arguments:
        * parent (ruamel.yaml node) The parent node from which this query
          originates
        * parentref (Any) The Index or Key of data within parent
        * relay_segment (PathSegment) YAML Path segment presently under
          evaluation
        * translated_path (YAMLPath) YAML Path indicating precisely which node
          is being evaluated
        * ancestry (List[AncestryEntry]) Stack of ancestors preceding the
          present node under evaluation

        Returns:  (Generator[NodeCoords, None, None]) each result as it is
            generated
        """
        # There must be exactly one parameter
        param_count = len(parameters)
        if param_count != 1:
            raise YAMLPathException(
                ("Invalid parameter count to {}; {} required, got {} in"
                 " YAML Path").format(
                     PathSearchKeywords.HAS_CHILD, 1, param_count),
                str(yaml_path))
        match_key = parameters[0]

        if match_key[0] == "&":
            matches = KeywordSearches._has_anchored_child(
                data, invert, parameters, yaml_path, **kwargs)
        else:
            matches = KeywordSearches._has_concrete_child(
                data, invert, parameters, yaml_path, **kwargs)

        for match in matches:
            yield match

    @staticmethod
    # pylint: disable=locally-disabled,too-many-locals
    def _has_concrete_child(
        data: Any, invert: bool, parameters: List[str], yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Indicate whether data has a named child.

        Parameters:
        1. data (Any) The data to evaluate
        2. invert (bool) Invert the evaluation
        3. parameters (List[str]) Parsed parameters
        4. yaml_path (YAMLPath) YAML Path begetting this operation

        Keyword Arguments:
        * parent (ruamel.yaml node) The parent node from which this query
          originates
        * parentref (Any) The Index or Key of data within parent
        * relay_segment (PathSegment) YAML Path segment presently under
          evaluation
        * translated_path (YAMLPath) YAML Path indicating precisely which node
          is being evaluated
        * ancestry (List[AncestryEntry]) Stack of ancestors preceding the
          present node under evaluation

        Returns:  (Generator[NodeCoords, None, None]) each result as it is
            generated
        """
        parent: Any = kwargs.pop("parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        translated_path: YAMLPath = kwargs.pop("translated_path", YAMLPath(""))
        ancestry: List[AncestryEntry] = kwargs.pop("ancestry", [])
        relay_segment: PathSegment = kwargs.pop("relay_segment", None)

        match_key = parameters[0]

        # Against a map, this will return nodes which have an immediate
        # child key exactly named as per parameters.  When inverted, only
        # parents with no such key are yielded.
        if isinstance(data, dict):
            child_present = data is not None and match_key in data
            if (
                (invert and not child_present) or
                (child_present and not invert)
            ):
                yield NodeCoords(
                    data, parent, parentref, translated_path, ancestry,
                    relay_segment)

        # Against a list, this will merely require an exact match between
        # parameters and any list elements.  When inverted, every
        # non-matching element is yielded.
        elif isinstance(data, list):
            # Against an AoH, this will scan each element's immediate children,
            # treating and yielding as if this search were performed directly
            # against each map in the list.
            if Nodes.node_is_aoh(data):
                for idx, ele in enumerate(data):
                    next_path = translated_path.append("[{}]".format(str(idx)))
                    for aoh_match in KeywordSearches._has_concrete_child(
                        ele, invert, parameters, yaml_path,
                        parent=data, parentref=idx, translated_path=next_path
                    ):
                        yield aoh_match
                return

            child_present = match_key in data
            if (
                (invert and not child_present) or
                (child_present and not invert)
            ):
                yield NodeCoords(
                    data, parent, parentref, translated_path, ancestry,
                    relay_segment)

        elif data is None:
            if invert:
                yield NodeCoords(
                    data, parent, parentref, translated_path, ancestry,
                    relay_segment)

        else:
            raise YAMLPathException(
                ("{} data has no child nodes in YAML Path").format(type(data)),
                str(yaml_path))

    @staticmethod
    # pylint: disable=locally-disabled,too-many-locals,too-many-branches,too-many-statements
    def _has_anchored_child(
        data: Any, invert: bool, parameters: List[str], yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Indicate whether data has an anchored child.

        Parameters:
        1. data (Any) The data to evaluate
        2. invert (bool) Invert the evaluation
        3. parameters (List[str]) Parsed parameters
        4. yaml_path (YAMLPath) YAML Path begetting this operation

        Keyword Arguments:
        * parent (ruamel.yaml node) The parent node from which this query
          originates
        * parentref (Any) The Index or Key of data within parent
        * relay_segment (PathSegment) YAML Path segment presently under
          evaluation
        * translated_path (YAMLPath) YAML Path indicating precisely which node
          is being evaluated
        * ancestry (List[AncestryEntry]) Stack of ancestors preceding the
          present node under evaluation

        Returns:  (Generator[NodeCoords, None, None]) each result as it is
            generated
        """
        parent: Any = kwargs.pop("parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        translated_path: YAMLPath = kwargs.pop("translated_path", YAMLPath(""))
        ancestry: List[AncestryEntry] = kwargs.pop("ancestry", [])
        relay_segment: PathSegment = kwargs.pop("relay_segment", None)

        match_key = parameters[0]
        anchor_name = match_key[1:] if match_key[0] == "&" else match_key

        if isinstance(data, CommentedMap):
            # Look for YAML Merge Keys by the Anchor name
            all_data = ancestry[0][0] if len(ancestry) > 0 else data
            all_anchors: Dict[str, Any] = {}
            Anchors.scan_for_anchors(all_data, all_anchors)
            compare_node = (all_anchors[anchor_name]
                            if anchor_name in all_anchors
                            else None)
            is_ymk_anchor = (
                compare_node is not None and isinstance(compare_node, dict))

            if is_ymk_anchor:
                child_present = False
                if hasattr(data, "merge") and len(data.merge) > 0:
                    # Ignore comparision if there is no source
                    for (idx, merge_node) in data.merge:
                        if merge_node == compare_node:
                            child_present = True
                            break

                if (
                    (invert and not child_present) or
                    (child_present and not invert)
                ):
                    yield NodeCoords(
                        data, parent, parentref, translated_path,
                        ancestry, relay_segment)
                    return

            # Look for Anchored keys; include merged nodes
            else:
                child_present = False
                for (key, val) in data.items():
                    key_anchor = Anchors.get_node_anchor(key)
                    val_anchor = Anchors.get_node_anchor(val)
                    if key_anchor and key_anchor == anchor_name:
                        child_present = True
                        break
                    if val_anchor and val_anchor == anchor_name:
                        child_present = True
                        break

                if (
                    (invert and not child_present) or
                    (child_present and not invert)
                ):
                    yield NodeCoords(
                        data, parent, parentref, translated_path,
                        ancestry, relay_segment)

        elif Nodes.node_is_aoh(data, accept_nulls=True):
            for idx, ele in enumerate(data):
                if ele is None:
                    continue

                next_path = translated_path.append("[{}]".format(str(idx)))
                next_ancestry = ancestry + [(data, idx)]
                for aoh_match in KeywordSearches._has_anchored_child(
                    ele, invert, parameters, yaml_path,
                    parent=data, parentref=idx, translated_path=next_path,
                    ancestry=next_ancestry
                ):
                    yield aoh_match

        elif isinstance(data, list):
            child_present = False
            for ele in data:
                ele_anchor = Anchors.get_node_anchor(ele)
                if ele_anchor and ele_anchor == anchor_name:
                    child_present = True
                    break

            if (
                (invert and not child_present) or
                (child_present and not invert)
            ):
                yield NodeCoords(
                    data, parent, parentref, translated_path,
                    ancestry, relay_segment)

    @staticmethod
    # pylint: disable=locally-disabled,too-many-locals
    def name(
        invert: bool, parameters: List[str], yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Match only the key-name of the present node.

        Parameters:
        1. invert (bool) Invert the evaluation
        2. parameters (List[str]) Parsed parameters
        3. yaml_path (YAMLPath) YAML Path begetting this operation

        Keyword Arguments:
        * parent (ruamel.yaml node) The parent node from which this query
          originates
        * parentref (Any) The Index or Key of data within parent
        * relay_segment (PathSegment) YAML Path segment presently under
          evaluation
        * translated_path (YAMLPath) YAML Path indicating precisely which node
          is being evaluated
        * ancestry (List[AncestryEntry]) Stack of ancestors preceding the
          present node under evaluation

        Returns:  (Generator[NodeCoords, None, None]) each result as it is
            generated
        """
        parent: Any = kwargs.pop("parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        translated_path: YAMLPath = kwargs.pop("translated_path", YAMLPath(""))
        ancestry: List[AncestryEntry] = kwargs.pop("ancestry", [])
        relay_segment: PathSegment = kwargs.pop("relay_segment", None)

        # There are no parameters
        param_count = len(parameters)
        if param_count > 1:
            raise YAMLPathException((
                "Invalid parameter count to {}(); {} are permitted, "
                " got {} in YAML Path"
                ).format(PathSearchKeywords.NAME, 0, param_count),
                str(yaml_path))

        if invert:
            raise YAMLPathException((
                "Inversion is meaningless to {}()"
                ).format(PathSearchKeywords.NAME),
                str(yaml_path))

        yield NodeCoords(
            parentref, parent, parentref, translated_path, ancestry,
            relay_segment)

    @staticmethod
    # pylint: disable=locally-disabled,too-many-locals,too-many-branches,too-many-statements
    def max(
        data: Any, invert: bool, parameters: List[str], yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Find whichever nodes/elements have a maximum value.

        Parameters:
        1. data (Any) The data to evaluate
        2. invert (bool) Invert the evaluation
        3. parameters (List[str]) Parsed parameters
        4. yaml_path (YAMLPath) YAML Path begetting this operation

        Keyword Arguments:
        * parent (ruamel.yaml node) The parent node from which this query
          originates
        * parentref (Any) The Index or Key of data within parent
        * relay_segment (PathSegment) YAML Path segment presently under
          evaluation
        * translated_path (YAMLPath) YAML Path indicating precisely which node
          is being evaluated
        * ancestry (List[AncestryEntry]) Stack of ancestors preceding the
          present node under evaluation

        Returns:  (Generator[NodeCoords, None, None]) each result as it is
            generated
        """
        parent: Any = kwargs.pop("parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        translated_path: YAMLPath = kwargs.pop("translated_path", YAMLPath(""))
        ancestry: List[AncestryEntry] = kwargs.pop("ancestry", [])
        relay_segment: PathSegment = kwargs.pop("relay_segment", None)

        # There may be 0 or 1 parameters
        param_count = len(parameters)
        if param_count > 1:
            raise YAMLPathException((
                "Invalid parameter count to {}([NAME]); up to {} permitted, "
                " got {} in YAML Path"
                ).format(PathSearchKeywords.MAX, 1, param_count),
                str(yaml_path))

        scan_node = parameters[0] if param_count > 0 else None
        match_value: Any = None
        match_nodes: List[NodeCoords] = []
        discard_nodes: List[NodeCoords] = []
        unwrapped_data: Any = NodeCoords.unwrap_node_coords(data)
        if Nodes.node_is_aoh(
            unwrapped_data, accept_nulls=True
        ):
            # A named child node is mandatory
            if scan_node is None:
                raise YAMLPathException((
                    "The {}([NAME]) Search Keyword requires a key name to scan"
                    " when evaluating an Array-of-Hashes in YAML Path"
                    ).format(PathSearchKeywords.MAX),
                    str(yaml_path))

            for idx, wrapped_ele in enumerate(data):
                ele = NodeCoords.unwrap_node_coords(wrapped_ele)
                next_path = translated_path + "[{}]".format(idx)
                next_ancestry = ancestry + [(data, idx)]
                if ele is not None and scan_node in ele:
                    eval_val = ele[scan_node]
                    if (match_value is None
                        or Searches.search_matches(
                            PathSearchMethods.GREATER_THAN, match_value,
                            eval_val)
                    ):
                        match_value = eval_val
                        discard_nodes.extend(match_nodes)
                        match_nodes = [
                            NodeCoords(
                                ele, data, idx, next_path, next_ancestry,
                                relay_segment)
                        ]
                        continue

                    if (match_value is None
                        or Searches.search_matches(
                            PathSearchMethods.EQUALS, match_value,
                            eval_val)
                    ):
                        match_nodes.append(NodeCoords(
                            ele, data, idx, next_path, next_ancestry,
                            relay_segment))
                        continue

                discard_nodes.append(NodeCoords(
                    ele, data, idx, next_path, next_ancestry,
                    relay_segment))

        elif isinstance(data, dict):
            # A named child node is mandatory
            if scan_node is None:
                raise YAMLPathException((
                    "The {}([NAME]) Search Keyword requires a key name to scan"
                    " when comparing Hash/map/dict children in YAML Path"
                    ).format(PathSearchKeywords.MAX),
                    str(yaml_path))

            for key, val in data.items():
                next_path = (
                    translated_path + YAMLPath.escape_path_section(
                        key, translated_path.seperator))
                next_ancestry = ancestry + [(data, key)]
                if isinstance(val, dict):
                    if val is not None and scan_node in val:
                        eval_val = val[scan_node]
                        if (match_value is None
                            or Searches.search_matches(
                                PathSearchMethods.GREATER_THAN, match_value,
                                eval_val)
                        ):
                            match_value = eval_val
                            discard_nodes.extend(match_nodes)
                            match_nodes = [
                                NodeCoords(
                                    val, data, key, next_path, next_ancestry,
                                    relay_segment)
                            ]
                            continue

                        if (match_value is None
                            or Searches.search_matches(
                                PathSearchMethods.EQUALS, match_value,
                                eval_val)
                        ):
                            match_nodes.append(NodeCoords(
                                val, data, key, next_path, next_ancestry,
                                relay_segment))
                            continue

                elif scan_node in data:
                    # The user probably meant to operate against the parent
                    raise YAMLPathException((
                        "The {}([NAME]) Search Keyword operates against"
                        " collections of data which share a common attribute"
                        " yet there is only a single node to consider.  Did"
                        " you mean to evaluate the parent of the selected"
                        " node?  Please review your YAML Path"
                        ).format(PathSearchKeywords.MAX),
                        str(yaml_path))

                discard_nodes.append(NodeCoords(
                    val, data, key, next_path, next_ancestry,
                    relay_segment))

        elif isinstance(data, list):
            # A named child node is useless
            if scan_node is not None:
                raise YAMLPathException((
                    "The {}([NAME]) Search Keyword cannot utilize a key name"
                    " when comparing Array/sequence/list elements to one"
                    " another in YAML Path"
                    ).format(PathSearchKeywords.MAX),
                    str(yaml_path))

            for idx, ele in enumerate(data):
                next_path = translated_path + "[{}]".format(idx)
                next_ancestry = ancestry + [(data, idx)]
                if (ele is not None
                    and (
                        match_value is None or
                        Searches.search_matches(
                            PathSearchMethods.GREATER_THAN, match_value,
                            ele)
                )):
                    match_value = ele
                    discard_nodes.extend(match_nodes)
                    match_nodes = [
                        NodeCoords(
                            ele, data, idx, next_path, next_ancestry,
                            relay_segment)
                    ]
                    continue

                if (ele is not None
                    and Searches.search_matches(
                        PathSearchMethods.EQUALS, match_value,
                        ele)
                ):
                    match_nodes.append(NodeCoords(
                        ele, data, idx, next_path, next_ancestry,
                        relay_segment))
                    continue

                discard_nodes.append(NodeCoords(
                    ele, data, idx, next_path, next_ancestry,
                    relay_segment))

        else:
            # Non-complex data is always its own maximum and does not invert
            match_value = data
            match_nodes = [
                NodeCoords(
                    data, parent, parentref, translated_path, ancestry,
                    relay_segment)
            ]

        yield_nodes = discard_nodes if invert else match_nodes
        for node_coord in yield_nodes:
            yield node_coord


    @staticmethod
    # pylint: disable=locally-disabled,too-many-locals,too-many-branches,too-many-statements
    def min(
        data: Any, invert: bool, parameters: List[str], yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Find whichever nodes/elements have a minimum value.

        Parameters:
        1. data (Any) The data to evaluate
        2. invert (bool) Invert the evaluation
        3. parameters (List[str]) Parsed parameters
        4. yaml_path (YAMLPath) YAML Path begetting this operation

        Keyword Arguments:
        * parent (ruamel.yaml node) The parent node from which this query
          originates
        * parentref (Any) The Index or Key of data within parent
        * relay_segment (PathSegment) YAML Path segment presently under
          evaluation
        * translated_path (YAMLPath) YAML Path indicating precisely which node
          is being evaluated
        * ancestry (List[AncestryEntry]) Stack of ancestors preceding the
          present node under evaluation

        Returns:  (Generator[NodeCoords, None, None]) each result as it is
            generated
        """
        parent: Any = kwargs.pop("parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        translated_path: YAMLPath = kwargs.pop("translated_path", YAMLPath(""))
        ancestry: List[AncestryEntry] = kwargs.pop("ancestry", [])
        relay_segment: PathSegment = kwargs.pop("relay_segment", None)

        # There may be 0 or 1 parameters
        param_count = len(parameters)
        if param_count > 1:
            raise YAMLPathException((
                "Invalid parameter count to {}([NAME]); up to {} permitted, "
                " got {} in YAML Path"
                ).format(PathSearchKeywords.MIN, 1, param_count),
                str(yaml_path))

        scan_node = parameters[0] if param_count > 0 else None
        match_value: Any = None
        match_nodes: List[NodeCoords] = []
        discard_nodes: List[NodeCoords] = []
        unwrapped_data: Any = NodeCoords.unwrap_node_coords(data)
        if Nodes.node_is_aoh(
            unwrapped_data, accept_nulls=True
        ):
            # A named child node is mandatory
            if scan_node is None:
                raise YAMLPathException((
                    "The {}([NAME]) Search Keyword requires a key name to scan"
                    " when evaluating an Array-of-Hashes in YAML Path"
                    ).format(PathSearchKeywords.MIN),
                    str(yaml_path))

            for idx, wrapped_ele in enumerate(data):
                ele = NodeCoords.unwrap_node_coords(wrapped_ele)
                next_path = translated_path + "[{}]".format(idx)
                next_ancestry = ancestry + [(data, idx)]
                if ele is not None and scan_node in ele:
                    eval_val = ele[scan_node]
                    if (match_value is None
                        or Searches.search_matches(
                            PathSearchMethods.LESS_THAN, match_value,
                            eval_val)
                    ):
                        match_value = eval_val
                        discard_nodes.extend(match_nodes)
                        match_nodes = [
                            NodeCoords(
                                ele, data, idx, next_path, next_ancestry,
                                relay_segment)
                        ]
                        continue

                    if (match_value is None
                        or Searches.search_matches(
                            PathSearchMethods.EQUALS, match_value,
                            eval_val)
                    ):
                        match_nodes.append(NodeCoords(
                            ele, data, idx, next_path, next_ancestry,
                            relay_segment))
                        continue

                discard_nodes.append(NodeCoords(
                    ele, data, idx, next_path, next_ancestry,
                    relay_segment))

        elif isinstance(data, dict):
            # A named child node is mandatory
            if scan_node is None:
                raise YAMLPathException((
                    "The {}([NAME]) Search Keyword requires a key name to scan"
                    " when comparing Hash/map/dict children in YAML Path"
                    ).format(PathSearchKeywords.MIN),
                    str(yaml_path))

            for key, val in data.items():
                next_ancestry = ancestry + [(data, key)]
                next_path = (
                    translated_path + YAMLPath.escape_path_section(
                        key, translated_path.seperator))
                if isinstance(val, dict):
                    if val is not None and scan_node in val:
                        eval_val = val[scan_node]
                        if (match_value is None
                            or Searches.search_matches(
                                PathSearchMethods.LESS_THAN, match_value,
                                eval_val)
                        ):
                            match_value = eval_val
                            discard_nodes.extend(match_nodes)
                            match_nodes = [
                                NodeCoords(
                                    val, data, key, next_path, next_ancestry,
                                    relay_segment)
                            ]
                            continue

                        if (match_value is None
                            or Searches.search_matches(
                                PathSearchMethods.EQUALS, match_value,
                                eval_val)
                        ):
                            match_nodes.append(NodeCoords(
                                val, data, key, next_path, next_ancestry,
                                relay_segment))
                            continue

                elif scan_node in data:
                    # The user probably meant to operate against the parent
                    raise YAMLPathException((
                        "The {}([NAME]) Search Keyword operates against"
                        " collections of data which share a common attribute"
                        " yet there is only a single node to consider.  Did"
                        " you mean to evaluate the parent of the selected"
                        " node?  Please review your YAML Path"
                        ).format(PathSearchKeywords.MIN),
                        str(yaml_path))

                discard_nodes.append(NodeCoords(
                    val, data, key, next_path, next_ancestry,
                    relay_segment))

        elif isinstance(data, list):
            # A named child node is useless
            if scan_node is not None:
                raise YAMLPathException((
                    "The {}([NAME]) Search Keyword cannot utilize a key name"
                    " when comparing Array/sequence/list elements to one"
                    " another in YAML Path"
                    ).format(PathSearchKeywords.MIN),
                    str(yaml_path))

            for idx, ele in enumerate(data):
                next_path = translated_path + "[{}]".format(idx)
                next_ancestry = ancestry + [(data, idx)]
                if (ele is not None
                    and (
                        match_value is None or
                        Searches.search_matches(
                            PathSearchMethods.LESS_THAN, match_value,
                            ele)
                )):
                    match_value = ele
                    discard_nodes.extend(match_nodes)
                    match_nodes = [
                        NodeCoords(
                            ele, data, idx, next_path, next_ancestry,
                            relay_segment)
                    ]
                    continue

                if (ele is not None
                    and Searches.search_matches(
                        PathSearchMethods.EQUALS, match_value,
                        ele)
                ):
                    match_nodes.append(NodeCoords(
                        ele, data, idx, next_path, next_ancestry,
                        relay_segment))
                    continue

                discard_nodes.append(NodeCoords(
                    ele, data, idx, next_path, next_ancestry,
                    relay_segment))

        else:
            # Non-complex data is always its own maximum and does not invert
            match_value = data
            match_nodes = [
                NodeCoords(
                    data, parent, parentref, translated_path, ancestry,
                    relay_segment)
            ]

        yield_nodes = discard_nodes if invert else match_nodes
        for node_coord in yield_nodes:
            yield node_coord

    @staticmethod
    # pylint: disable=locally-disabled,too-many-locals
    def parent(
        data: Any, invert: bool, parameters: List[str], yaml_path: YAMLPath,
        **kwargs: Any
    ) -> Generator[NodeCoords, None, None]:
        """
        Climb back up N parent levels in the data hierarchy.

        Parameters:
        1. data (Any) The data to evaluate
        2. invert (bool) Invert the evaluation; not possible for parent()
        3. parameters (List[str]) Parsed parameters
        4. yaml_path (YAMLPath) YAML Path begetting this operation

        Keyword Arguments:
        * parent (ruamel.yaml node) The parent node from which this query
          originates
        * parentref (Any) The Index or Key of data within parent
        * relay_segment (PathSegment) YAML Path segment presently under
          evaluation
        * translated_path (YAMLPath) YAML Path indicating precisely which node
          is being evaluated
        * ancestry (List[AncestryEntry]) Stack of ancestors preceding the
          present node under evaluation

        Returns:  (Generator[NodeCoords, None, None]) each result as it is
            generated
        """
        parent: Any = kwargs.pop("parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        translated_path: YAMLPath = kwargs.pop("translated_path", YAMLPath(""))
        ancestry: List[AncestryEntry] = kwargs.pop("ancestry", [])
        relay_segment: PathSegment = kwargs.pop("relay_segment", None)

        # There may be 0 or 1 parameters
        param_count = len(parameters)
        if param_count > 1:
            raise YAMLPathException((
                "Invalid parameter count to {}([STEPS]); up to {} permitted, "
                " got {} in YAML Path"
                ).format(PathSearchKeywords.PARENT, 1, param_count),
                str(yaml_path))

        if invert:
            raise YAMLPathException((
                "Inversion is meaningless to {}([STEPS])"
                ).format(PathSearchKeywords.PARENT),
                str(yaml_path))

        parent_levels: int = 1
        ancestry_len: int = len(ancestry)
        steps_max = ancestry_len
        if param_count > 0:
            try:
                parent_levels = int(parameters[0])
            except ValueError as ex:
                raise YAMLPathException((
                    "Invalid parameter passed to {}([STEPS]), {}; must be"
                    " unset or an integer number indicating how may parent"
                    " STEPS to climb in YAML Path"
                    ).format(PathSearchKeywords.PARENT, parameters[0]),
                    str(yaml_path)) from ex

        if parent_levels > steps_max:
            raise YAMLPathException((
                "Cannot {}([STEPS]) higher than the document root.  {} steps"
                " requested when {} available in YAML Path"
                ).format(PathSearchKeywords.PARENT, parent_levels, steps_max),
                str(yaml_path))

        if parent_levels < 1:
            # parent(0) is the present node
            yield NodeCoords(
                data, parent, parentref, translated_path, ancestry,
                relay_segment)
        else:
            for _ in range(parent_levels):
                translated_path.pop()
                (data, _) = ancestry.pop()
                ancestry_len -= 1

            parentref = ancestry[-1][1] if ancestry_len > 0 else None
            parent = ancestry[-1][0] if ancestry_len > 0 else None
            yield NodeCoords(
                data, parent, parentref, translated_path, ancestry,
                relay_segment)



"""
Implements the PathSearchMethods enumeration.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto
from typing import List


class PathSearchMethods(Enum):
    """
    Supported methods for searching YAML Path segments.

    These include:

    `CONTAINS`
        Matches when the haystack contains the needle.

    `ENDS_WITH`
        Matches when the haystack ends with the needle.

    `EQUALS`
        Matches when the haystack and needle are identical.

    `STARTS_WITH`
        Matches when the haystack starts with the needle.

    `GREATER_THAN`
        Matches when the needle is greater than the haystack.

    `LESS_THAN`
        Matches when the needle is less than the haystack.

    `GREATER_THAN_OR_EQUAL`
        Matches when the needle is greater than or equal to the haystack.

    `LESS_THAN_OR_EQUAL`
        Matches when the needle is less than or equal to the haystack.

    `REGEX`
        Matches when the needle Regular Expression matches the haystack.
    """

    CONTAINS = auto()
    ENDS_WITH = auto()
    EQUALS = auto()
    STARTS_WITH = auto()
    GREATER_THAN = auto()
    LESS_THAN = auto()
    GREATER_THAN_OR_EQUAL = auto()
    LESS_THAN_OR_EQUAL = auto()
    REGEX = auto()

    def __str__(self) -> str:
        """Get a String representation of an employed value of this enum."""
        operator = ''
        if self is PathSearchMethods.EQUALS:
            operator = '='
        elif self is PathSearchMethods.STARTS_WITH:
            operator = '^'
        elif self is PathSearchMethods.ENDS_WITH:
            operator = '$'
        elif self is PathSearchMethods.CONTAINS:
            operator = '%'
        elif self is PathSearchMethods.LESS_THAN:
            operator = '<'
        elif self is PathSearchMethods.GREATER_THAN:
            operator = '>'
        elif self is PathSearchMethods.LESS_THAN_OR_EQUAL:
            operator = '<='
        elif self is PathSearchMethods.GREATER_THAN_OR_EQUAL:
            operator = '>='
        elif self is PathSearchMethods.REGEX:
            operator = '=~'

        return operator

    @staticmethod
    def get_operators() -> List[str]:
        """Return the full list of supported symbolic search operators."""
        return [str(o) for o in PathSearchMethods]

    @staticmethod
    def is_operator(symbol: str) -> bool:
        """Indicate whether symbol is a known search method operator."""
        return symbol in PathSearchMethods.get_operators()



import pytest

from yamlpath.enums import PathSeperators


class Test_enums_PathSeperators():
	"""Tests for the PathSeperators enumeration."""
	def test_get_names(self):
		assert PathSeperators.get_names() == [
			"AUTO",
			"DOT",
			"FSLASH",
		]

	@pytest.mark.parametrize("input,output", [
		(PathSeperators.AUTO, '.'),
		(PathSeperators.DOT, '.'),
		(PathSeperators.FSLASH, '/'),
	])
	def test_str(self, input, output):
		assert output == str(input)

	@pytest.mark.parametrize("input,output", [
		(".", PathSeperators.DOT),
		("/", PathSeperators.FSLASH),
		("DOT", PathSeperators.DOT),
		("FSLASH", PathSeperators.FSLASH),
		(PathSeperators.DOT, PathSeperators.DOT),
		(PathSeperators.FSLASH, PathSeperators.FSLASH),
	])
	def test_from_str(self, input, output):
		assert output == PathSeperators.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			PathSeperators.from_str("NO SUCH NAME")

	@pytest.mark.parametrize("input,output", [
		("abc", PathSeperators.DOT),
		("abc.123", PathSeperators.DOT),
		("/abc", PathSeperators.FSLASH),
		("/abc/123", PathSeperators.FSLASH),
	])
	def test_infer_seperator(self, input, output):
		assert output == PathSeperators.infer_seperator(input)



"""
Implements the AoHDiffOpts enumeration.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto
from typing import List


class AoHDiffOpts(Enum):
    """
    Supported Array-of-Hash (AKA: List-of-Dictionaries) Diff Options.

    Options include:

    `DEEP`
        Like KEY except the record pairs are deeply traversed, looking for
        specific internal differences, after being matched up.

    `DPOS`
        Like POSITION (no KEY matching) except the record pairs are deeply
        traversed to report every specific difference between them.

    `KEY`
        AoH records are synchronized by their identity key before being
        compared as whole units (no deep traversal).

    `POSITION`
        AoH records are compared as whole units (no deep traversal) based on
        their ordinal position in each document.

    `VALUE`
        AoH records are synchronized as whole units (no deep traversal) before
        being compared.
    """

    DEEP = auto()
    DPOS = auto()
    KEY = auto()
    POSITION = auto()
    VALUE = auto()

    def __str__(self) -> str:
        """
        Stringify one instance of this enumeration.

        Parameters:  N/A

        Returns:  (str) String value of this enumeration.

        Raises:  N/A
        """
        return str(self.name).lower()

    @staticmethod
    def get_names() -> List[str]:
        """
        Get all upper-cased entry names for this enumeration.

        Parameters:  N/A

        Returns:  (List[str]) Upper-case names from this enumeration

        Raises:  N/A
        """
        return [entry.name.upper() for entry in AoHDiffOpts]

    @staticmethod
    def get_choices() -> List[str]:
        """
        Get all entry names with symbolic representations for this enumeration.

        All returned entries are lower-cased.

        Parameters:  N/A

        Returns:  (List[str]) Lower-case names and symbols from this
            enumeration

        Raises:  N/A
        """
        names = [l.lower() for l in AoHDiffOpts.get_names()]
        choices = list(set(names))
        choices.sort()
        return choices

    @staticmethod
    def from_str(name: str) -> "AoHDiffOpts":
        """
        Convert a string value to a value of this enumeration, if valid.

        Parameters:
            1. name (str) The name to convert

        Returns:  (AoHDiffOpts) the converted enumeration value

        Raises:
            - `NameError` when name doesn't match any enumeration values
        """
        check: str = str(name).upper()
        if check in AoHDiffOpts.get_names():
            return AoHDiffOpts[check]
        raise NameError(
            "AoHDiffOpts has no such item:  {}"
            .format(name))



"""
Implements the AnchorConflictResolutions enumeration.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto
from typing import List


class AnchorConflictResolutions(Enum):
    """
    Supported Anchor Conflict Resolutions.

    Resolutions include:

    `STOP`
        Abort the merge upon conflict detection.

    `LEFT`
        The first-encountered definition overrides all other uses.

    `RIGHT`
        The last-encountered definition overrides all other uses.

    `RENAME`
        Conflicting anchors are renamed within the affected documents.
    """

    STOP = auto()
    LEFT = auto()
    RIGHT = auto()
    RENAME = auto()

    @staticmethod
    def get_names() -> List[str]:
        """
        Get all upper-cased entry names for this enumeration.

        Parameters:  N/A

        Returns:  (List[str]) Upper-case names from this enumeration

        Raises:  N/A
        """
        return [entry.name.upper() for entry in AnchorConflictResolutions]

    @staticmethod
    def get_choices() -> List[str]:
        """
        Get all entry names with symbolic representations for this enumeration.

        All returned entries are lower-cased.

        Parameters:  N/A

        Returns:  (List[str]) Lower-case names and symbols from this
            enumeration

        Raises:  N/A
        """
        names = [l.lower() for l in AnchorConflictResolutions.get_names()]
        choices = list(set(names))
        choices.sort()
        return choices

    @staticmethod
    def from_str(name: str) -> "AnchorConflictResolutions":
        """
        Convert a string value to a value of this enumeration, if valid.

        Parameters:
            1. name (str) The name to convert

        Returns:  (AnchorConflictResolutions) the converted enumeration value

        Raises:
            - `NameError` when name doesn't match any enumeration values
        """
        check: str = str(name).upper()
        if check in AnchorConflictResolutions.get_names():
            return AnchorConflictResolutions[check]
        raise NameError(
            "AnchorConflictResolutions has no such item:  {}"
            .format(name))



import pytest

from yamlpath.eyaml.enums import EYAMLOutputFormats


class Test_eyaml_enums_EYAMLOutputFormats():
	def test_get_names(self):
		assert EYAMLOutputFormats.get_names() == [
			"BLOCK",
			"STRING",
		]

	@pytest.mark.parametrize("input,output", [
		(EYAMLOutputFormats.BLOCK, "block"),
		(EYAMLOutputFormats.STRING, "string"),
	])
	def test_str(self, input, output):
		assert output == str(input)

	@pytest.mark.parametrize("input,output", [
		("block", EYAMLOutputFormats.BLOCK),
		("string", EYAMLOutputFormats.STRING),
		("BLOCK", EYAMLOutputFormats.BLOCK),
		("STRING", EYAMLOutputFormats.STRING),
		(EYAMLOutputFormats.BLOCK, EYAMLOutputFormats.BLOCK),
		(EYAMLOutputFormats.STRING, EYAMLOutputFormats.STRING),
	])
	def test_from_str(self, input, output):
		assert output == EYAMLOutputFormats.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			EYAMLOutputFormats.from_str("NO SUCH NAME")



"""
Implement MergerConfig.

Copyright 2020, 2021 William W. Kimball, Jr. MBA MSIS
"""
import configparser
from typing import Any, Dict, Union
from argparse import Namespace

from yamlpath.exceptions import YAMLPathException
from yamlpath.merger.enums import (
    AnchorConflictResolutions,
    AoHMergeOpts,
    ArrayMergeOpts,
    HashMergeOpts,
    MultiDocModes,
    OutputDocTypes,
    SetMergeOpts,
)
from yamlpath import Processor, YAMLPath
from yamlpath.wrappers import ConsolePrinter, NodeCoords


class MergerConfig:
    """Config file processor for the Merger."""

    def __init__(self, logger: ConsolePrinter, args: Namespace) -> None:
        """
        Instantiate this class into an object.

        Parameters:
        1. logger (ConsolePrinter) Instance of ConsoleWriter or subclass
        2. args (dict) Default options for merge rules

        Returns:  N/A
        """
        self.log = logger
        self.args = args
        self.config: Union[None, configparser.ConfigParser] = None
        self.rules: Dict[NodeCoords, str] = {}
        self.keys: Dict[NodeCoords, str] = {}

        self._load_config()

    def anchor_merge_mode(self) -> AnchorConflictResolutions:
        """
        Get Anchor merge mode.

        Parameters:  N/A

        Returns:  (AnchorConflictResolutions) Resolved method for handling
            YAML Anchor conflicts.
        """
        # Precedence: CLI > config[defaults] > default
        if hasattr(self.args, "anchors") and self.args.anchors:
            return AnchorConflictResolutions.from_str(self.args.anchors)
        if (self.config is not None
                and "defaults" in self.config
                and "anchors" in self.config["defaults"]):
            return AnchorConflictResolutions.from_str(
                self.config["defaults"]["anchors"])
        return AnchorConflictResolutions.STOP

    def hash_merge_mode(self, node_coord: NodeCoords) -> HashMergeOpts:
        """
        Get Hash merge mode applicable to the indicated path.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.

        Returns:  (HashMergeOpts) Applicable mode.
        """
        # Precedence: config[rules] > CLI > config[defaults] > default
        merge_rule = self._get_rule_for(node_coord)
        if merge_rule:
            self.log.debug(
                "MergerConfig::hash_merge_mode:  Matched {}"
                .format(merge_rule))
            return HashMergeOpts.from_str(merge_rule)
        self.log.debug("MergerConfig::hash_merge_mode:  NOT Matched")
        if hasattr(self.args, "hashes") and self.args.hashes:
            return HashMergeOpts.from_str(self.args.hashes)
        if (self.config is not None
                and "defaults" in self.config
                and "hashes" in self.config["defaults"]):
            return HashMergeOpts.from_str(self.config["defaults"]["hashes"])
        return HashMergeOpts.DEEP

    def array_merge_mode(self, node_coord: NodeCoords) -> ArrayMergeOpts:
        """
        Get Array merge mode applicable to the indicated path.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.

        Returns:  (ArrayMergeOpts) Applicable mode.
        """
        # Precedence: config[rules] > CLI > config[defaults] > default
        merge_rule = self._get_rule_for(node_coord)
        if merge_rule:
            self.log.debug(
                "MergerConfig::array_merge_mode:  Matched {}"
                .format(merge_rule))
            return ArrayMergeOpts.from_str(merge_rule)
        self.log.debug("MergerConfig::array_merge_mode:  NOT Matched")
        if hasattr(self.args, "arrays") and self.args.arrays:
            return ArrayMergeOpts.from_str(self.args.arrays)
        if (self.config is not None
                and "defaults" in self.config
                and "arrays" in self.config["defaults"]):
            return ArrayMergeOpts.from_str(self.config["defaults"]["arrays"])
        return ArrayMergeOpts.ALL

    def aoh_merge_mode(self, node_coord: NodeCoords) -> AoHMergeOpts:
        """
        Get Array-of-Hashes merge mode applicable to the indicated path.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.

        Returns:  (AoHMergeOpts) Applicable mode.
        """
        # Precedence: config[rules] > CLI > config[defaults] > default
        merge_rule = self._get_rule_for(node_coord)
        if merge_rule:
            self.log.debug(
                "MergerConfig::aoh_merge_mode:  Matched {}"
                .format(merge_rule))
            return AoHMergeOpts.from_str(merge_rule)
        self.log.debug("MergerConfig::aoh_merge_mode:  NOT Matched")
        if hasattr(self.args, "aoh") and self.args.aoh:
            return AoHMergeOpts.from_str(self.args.aoh)
        if (self.config is not None
                and "defaults" in self.config
                and "aoh" in self.config["defaults"]):
            return AoHMergeOpts.from_str(self.config["defaults"]["aoh"])
        return AoHMergeOpts.ALL

    def set_merge_mode(self, node_coord: NodeCoords) -> SetMergeOpts:
        """
        Get Set merge mode applicable to the indicated path.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.

        Returns:  (SetMergeOpts) Applicable mode.
        """
        # Precedence: config[rules] > CLI > config[defaults] > default
        merge_rule = self._get_rule_for(node_coord)
        if merge_rule:
            self.log.debug(
                "MergerConfig::set_merge_mode:  Matched {}"
                .format(merge_rule))
            return SetMergeOpts.from_str(merge_rule)
        self.log.debug("MergerConfig::set_merge_mode:  NOT Matched")
        if hasattr(self.args, "sets") and self.args.sets:
            return SetMergeOpts.from_str(self.args.sets)
        if (self.config is not None
                and "defaults" in self.config
                and "sets" in self.config["defaults"]):
            return SetMergeOpts.from_str(self.config["defaults"]["sets"])
        return SetMergeOpts.UNIQUE

    def aoh_merge_key(
        self, node_coord: NodeCoords, data: dict
    ) -> str:
        """
        Get the user-defined identity key for Array-of-Hashes merging.

        Parameters:
        1. node_coord (NodeCoords) The node for which to query.
        2. data (dict) The merge source node from which an identity key will
           be inferred if not explicity provided via user configuration.

        Returns: (str) The identity key field name.
        """
        # Check the user config for a specific key; fallback to first key.
        merge_key = self._get_key_for(node_coord)
        if not merge_key:
            # This node may be a child of one of the registered keys.  That
            # registered key's node will match this node's parent.
            for eval_nc, eval_key in self.keys.items():
                if node_coord.parent == eval_nc.node:
                    merge_key = eval_key
                    break
        if not merge_key and len(data.keys()) > 0:
            # Fallback to using the first key of the dict as an identity key
            merge_key = list(data)[0]
        return merge_key

    def prepare(self, data: Any) -> None:
        """
        Load references to all nodes which match config rules.

        Parameters:
        1. data (Any) The DOM for which to load configuration.

        Returns:  N/A
        """
        if self.config is None:
            return

        # Eliminate previous rules and keys to limit scanning to only those
        # nodes which exist within this new document.
        self.rules = {}
        self.keys = {}

        # Load new rules and keys
        merge_path = self.get_insertion_point()
        proc = Processor(self.log, data)
        self._prepare_user_rules(proc, merge_path, "rules", self.rules)
        self._prepare_user_rules(proc, merge_path, "keys", self.keys)

    def get_insertion_point(self) -> YAMLPath:
        """
        Get the YAML Path at which merging shall be performed.

        Parameters:  N/A

        Returns:  (YAMLPath) User-specified point(s) within the document where
            the RHS document is directed to be merged-in.
        """
        if hasattr(self.args, "mergeat"):
            return YAMLPath(self.args.mergeat)
        return YAMLPath("/")

    def get_document_format(self) -> OutputDocTypes:
        """
        Get the user-desired output format.

        Paramerers:  N/A

        Returns:  (OutputDocTypes) The destination document type
        """
        if hasattr(self.args, "document_format"):
            return OutputDocTypes.from_str(self.args.document_format)
        return OutputDocTypes.AUTO

    def get_multidoc_mode(self) -> MultiDocModes:
        """
        Get the user-desired mode for handling multi-document merges.

        Paramerers:  N/A

        Returns:  (MultiDocModes) The condence/merge mode
        """
        if hasattr(self.args, "multi_doc_mode"):
            return MultiDocModes.from_str(self.args.multi_doc_mode)
        return MultiDocModes.CONDENSE_ALL

    def is_preserving_lhs_comments(self) -> bool:
        """Indicate whether the user wants LHS comments preserved."""
        if hasattr(self.args, "preserve_lhs_comments"):
            is_preserving: bool = self.args.preserve_lhs_comments
            return is_preserving
        return False

    def _prepare_user_rules(
        self, proc: Processor, merge_path: YAMLPath, section: str,
        collector: dict
    ) -> None:
        """
        Identify DOM nodes matching user-defined merge rules.

        Parameters:
        1. proc (Processor) Reference to the DOM Processor.
        2. merge_path (YAMLPath) User-specified path within the DOM at which
           merging will take place.
        3. section (str) User-configuration file section defining the merge
           rules to apply.
        4. collector (dict) Storage collector for matching nodes.

        Returns:  N/A
        """
        if self.config is None or not section in self.config:
            self.log.warning(
                "User-specified configuration file has no {} section."
                .format(section))
            return

        for rule_key in self.config[section]:
            rule_value = self.config[section][rule_key]

            if "=" in rule_value:
                # There were at least two = signs on the configuration line
                conf_line = rule_key + "=" + rule_value
                delim_pos = conf_line.rfind("=")
                rule_key = conf_line[0:delim_pos].strip()
                rule_value = conf_line[delim_pos + 1:].strip()
                self.log.debug(
                    "MergerConfig::_prepare_user_rules:  Reconstituted"
                    " configuration line '{}' to extract adjusted key '{}'"
                    " with value '{}'".format(conf_line, rule_key, rule_value))

            rule_path = YAMLPath(rule_key)
            yaml_path = YAMLPath.strip_path_prefix(rule_path, merge_path)
            self.log.debug(
                "MergerConfig::_prepare_user_rules:  Matching '{}' nodes to"
                " YAML Path '{}' from key, {}."
                .format(section, yaml_path, rule_key))
            try:
                for node_coord in proc.get_nodes(yaml_path, mustexist=True):
                    self.log.debug(
                        "Node will have merging rule, {}:"
                        .format(rule_value),
                        prefix="MergerConfig::_prepare_user_rules:  ",
                        data=node_coord.node)
                    collector[node_coord] = rule_value

            except YAMLPathException:
                self.log.warning("{} YAML Path matches no nodes:  {}"
                                .format(section, yaml_path))

        self.log.debug(
            "Matched rules to nodes:",
            prefix="MergerConfig::_prepare_user_rules:  ")
        for node_coord, merge_rule in collector.items():
            self.log.debug(
                "... RULE:  {}".format(merge_rule),
                prefix="MergerConfig::_prepare_user_rules:  ")
            self.log.debug(
                "... NODE:", data=node_coord,
                prefix="MergerConfig::_prepare_user_rules:  ")

    def _load_config(self) -> None:
        """Load the external configuration file."""
        config = configparser.ConfigParser()

        # Load the configuration file when one is specified
        config_file = (
            self.args.config
            if hasattr(self.args, "config")
            else None)

        if config_file:
            config.read(config_file)
            if config.sections():
                self.config = config

    def _get_config_for(self, node_coord: NodeCoords, section: dict) -> str:
        """
        Get user configuration applicable to a node.

        Parameters:
        1. node_coord (NodeCoords) The node for which to retrieve config.
        2. section (dict) The configuration section to query.

        Returns: (str) The requested configuration.
        """
        if self.config is None:
            return ""

        for rule_coord, rule_config in section.items():
            if rule_coord.node == node_coord.node \
                    and rule_coord.parent == node_coord.parent \
                    and rule_coord.parentref == node_coord.parentref:
                return str(rule_config)

        return ""

    def _get_rule_for(self, node_coord: NodeCoords) -> str:
        """
        Get a user configured merge rule for a node.

        Parameters:
        1. node_coord (NodeCoords) The node for which to retrieve config.

        Returns: (str) The requested configuration.
        """
        self.log.debug(
            "Seeking rule for node:", prefix="MergerConfig::_get_rule_for:  ",
            header=" ")
        self.log.debug(
            "... NODE:", prefix="MergerConfig::_get_rule_for:  ",
            data=node_coord)
        return self._get_config_for(node_coord, self.rules)

    def _get_key_for(self, node_coord: NodeCoords) -> str:
        """
        Get a user configured merge identity key (field) for a node.

        Parameters:
        1. node_coord (NodeCoords) The node for which to retrieve config.

        Returns: (str) The requested configuration.
        """
        return self._get_config_for(node_coord, self.keys)



import pytest

from yamlpath.wrappers import NodeCoords

class Test_wrappers_NodeCoords():
	"""Tests for the NodeCoords class."""

	def test_generic(self):
		node_coord = NodeCoords([], None, None)

	def test_repr(self):
		node_coord = NodeCoords([], None, None)
		assert repr(node_coord) == "NodeCoords('[]', 'None', 'None')"

	def test_str(self):
		node_coord = NodeCoords([], None, None)
		assert str(node_coord) == "[]"

	def test_gt(self):
		lhs_nc = NodeCoords(5, None, None)
		rhs_nc = NodeCoords(3, None, None)
		assert lhs_nc > rhs_nc

	def test_null_gt(self):
		lhs_nc = NodeCoords(5, None, None)
		rhs_nc = NodeCoords(None, None, None)
		assert not lhs_nc > rhs_nc

	def test_lt(self):
		lhs_nc = NodeCoords(5, None, None)
		rhs_nc = NodeCoords(7, None, None)
		assert lhs_nc < rhs_nc

	def test_null_lt(self):
		lhs_nc = NodeCoords(5, None, None)
		rhs_nc = NodeCoords(None, None, None)
		assert not lhs_nc < rhs_nc

	def test_isa_null(self):
		nc = NodeCoords(None, None, None)
		assert nc.wraps_a(None)
